import tensorflow as tf
from tensorflow import keras
from SamplingLayer import SamplingLayer
from utils import WINDOW_SIZE

Fs = 100

def model_baseline_2019(n_classes=5, use_sub_layer=False, use_rnn=True):
    """Recurrent_Deep_Neural_Networks_for_Real-Time_Sleep
    """
    inputLayer = tf.keras.layers.Input(shape=(3000, 1), name='inLayer')
    convFine = tf.keras.layers.Conv1D(filters=64, kernel_size=int(Fs/2), strides=int(Fs/16), padding='same', activation='relu', name='fConv1')(inputLayer)
    convFine = tf.keras.layers.MaxPool1D(pool_size=8, strides=8, name='fMaxP1')(convFine)
    convFine = tf.keras.layers.Dropout(rate=0.5, name='fDrop1')(convFine)
    convFine = tf.keras.layers.Conv1D(filters=128, kernel_size=8, padding='same', activation='relu', name='fConv2')(convFine)
    convFine = tf.keras.layers.Conv1D(filters=128, kernel_size=8, padding='same', activation='relu', name='fConv3')(convFine)
    convFine = tf.keras.layers.Conv1D(filters=128, kernel_size=8, padding='same', activation='relu', name='fConv4')(convFine)
    convFine = tf.keras.layers.MaxPool1D(pool_size=4, strides=4, name='fMaxP2')(convFine)
    fineShape = convFine.get_shape()
    convFine = tf.keras.layers.Flatten(name='fFlat1')(convFine)
    
    # network to learn coarse features
    convCoarse = tf.keras.layers.Conv1D(filters=32, kernel_size=Fs*4, strides=int(Fs/2), padding='same', activation='relu', name='cConv1')(inputLayer)
    convCoarse = tf.keras.layers.MaxPool1D(pool_size=4, strides=4, name='cMaxP1')(convCoarse)
    convCoarse = tf.keras.layers.Dropout(rate=0.5, name='cDrop1')(convCoarse)
    convCoarse = tf.keras.layers.Conv1D(filters=128, kernel_size=6, padding='same', activation='relu', name='cConv2')(convCoarse)
    convCoarse = tf.keras.layers.Conv1D(filters=128, kernel_size=6, padding='same', activation='relu', name='cConv3')(convCoarse)
    convCoarse = tf.keras.layers.Conv1D(filters=128, kernel_size=6, padding='same', activation='relu', name='cConv4')(convCoarse)
    convCoarse = tf.keras.layers.MaxPool1D(pool_size=2, strides=2, name='cMaxP2')(convCoarse)
    coarseShape = convCoarse.get_shape()
    convCoarse = tf.keras.layers.Flatten(name='cFlat1')(convCoarse)
    
    # concatenate coarse and fine cnns
    mergeLayer = tf.keras.layers.concatenate([convFine, convCoarse], name='merge_1')
    outLayer = tf.keras.layers.Dropout(rate=0.5, name='mDrop1')(mergeLayer)
    
    outLayer = tf.keras.layers.Reshape((1, outLayer.get_shape()[1]), name='reshape1')(outLayer)
    outLayer = tf.keras.layers.LSTM(64, return_sequences=True)(outLayer)
    outLayer = tf.keras.layers.LSTM(64, return_sequences=False)(outLayer)

    # Classify
    outLayer = tf.keras.layers.Dense(n_classes, activation='softmax', name='outLayer')(outLayer)
    model = tf.keras.Model(inputLayer, outLayer)
    optimizer = keras.optimizers.Adam(lr=1e-4)
    model.compile(optimizer=optimizer, loss='categorical_crossentropy', metrics=['acc'])
    model.summary()
    return model


def model_baseline_2019_sampling(n_classes=5):
    """Recurrent_Deep_Neural_Networks_for_Real-Time_Sleep
    """
    inputLayer = tf.keras.layers.Input(shape=(3000, 1), name='inLayer')
    samplingLayer = SamplingLayer(input_dim=3000)(inputLayer)
    convFine = tf.keras.layers.Conv1D(filters=64, kernel_size=int(Fs/2), strides=int(Fs/16), padding='same', activation='relu', name='fConv1')(samplingLayer)
    convFine = tf.keras.layers.MaxPool1D(pool_size=8, strides=8, name='fMaxP1')(convFine)
    convFine = tf.keras.layers.Dropout(rate=0.5, name='fDrop1')(convFine)
    convFine = tf.keras.layers.Conv1D(filters=128, kernel_size=8, padding='same', activation='relu', name='fConv2')(convFine)
    convFine = tf.keras.layers.Conv1D(filters=128, kernel_size=8, padding='same', activation='relu', name='fConv3')(convFine)
    convFine = tf.keras.layers.Conv1D(filters=128, kernel_size=8, padding='same', activation='relu', name='fConv4')(convFine)
    convFine = tf.keras.layers.MaxPool1D(pool_size=4, strides=4, name='fMaxP2')(convFine)
    fineShape = convFine.get_shape()
    convFine = tf.keras.layers.Flatten(name='fFlat1')(convFine)
    
    # network to learn coarse features
    convCoarse = tf.keras.layers.Conv1D(filters=32, kernel_size=Fs*4, strides=int(Fs/2), padding='same', activation='relu', name='cConv1')(samplingLayer)
    convCoarse = tf.keras.layers.MaxPool1D(pool_size=4, strides=4, name='cMaxP1')(convCoarse)
    convCoarse = tf.keras.layers.Dropout(rate=0.5, name='cDrop1')(convCoarse)
    convCoarse = tf.keras.layers.Conv1D(filters=128, kernel_size=6, padding='same', activation='relu', name='cConv2')(convCoarse)
    convCoarse = tf.keras.layers.Conv1D(filters=128, kernel_size=6, padding='same', activation='relu', name='cConv3')(convCoarse)
    convCoarse = tf.keras.layers.Conv1D(filters=128, kernel_size=6, padding='same', activation='relu', name='cConv4')(convCoarse)
    convCoarse = tf.keras.layers.MaxPool1D(pool_size=2, strides=2, name='cMaxP2')(convCoarse)
    coarseShape = convCoarse.get_shape()
    convCoarse = tf.keras.layers.Flatten(name='cFlat1')(convCoarse)
    
    # concatenate coarse and fine cnns
    mergeLayer = tf.keras.layers.concatenate([convFine, convCoarse], name='merge_1')
    outLayer = tf.keras.layers.Dropout(rate=0.5, name='mDrop1')(mergeLayer)
    
    outLayer = tf.keras.layers.Reshape((1, outLayer.get_shape()[1]), name='reshape1')(outLayer)
    outLayer = tf.keras.layers.LSTM(64, return_sequences=True)(outLayer)
    outLayer = tf.keras.layers.LSTM(64, return_sequences=False)(outLayer)

    # Classify
    outLayer = tf.keras.layers.Dense(n_classes, activation='softmax', name='outLayer')(outLayer)
    model = tf.keras.Model(inputLayer, outLayer)
    optimizer = keras.optimizers.Adam(lr=1e-4)
    model.compile(optimizer=optimizer, loss='categorical_crossentropy', metrics=['acc'])
    model.summary()
    return model


def define_sampling_subnet():
    seq_input = tf.keras.layers.Input(shape=(3000, 1))
    sampling_layer = SamplingLayer(input_dim=3000)(seq_input)
    model = tf.keras.models.Model(seq_input, sampling_layer)
    model.compile(tf.keras.optimizers.Adam(0.001), tf.keras.losses.categorical_crossentropy, metrics=['acc'])
    model.summary()
    return model

def define_inference_subnet(n_classes=5):
    inputLayer = tf.keras.layers.Input(shape=(3000, 1), name='inLayer')
    convFine = tf.keras.layers.Conv1D(filters=64, kernel_size=int(Fs/2), strides=int(Fs/16), padding='same', activation='relu', name='fConv1')(inputLayer)
    convFine = tf.keras.layers.MaxPool1D(pool_size=8, strides=8, name='fMaxP1')(convFine)
    convFine = tf.keras.layers.Dropout(rate=0.5, name='fDrop1')(convFine)
    convFine = tf.keras.layers.Conv1D(filters=128, kernel_size=8, padding='same', activation='relu', name='fConv2')(convFine)
    convFine = tf.keras.layers.Conv1D(filters=128, kernel_size=8, padding='same', activation='relu', name='fConv3')(convFine)
    convFine = tf.keras.layers.Conv1D(filters=128, kernel_size=8, padding='same', activation='relu', name='fConv4')(convFine)
    convFine = tf.keras.layers.MaxPool1D(pool_size=4, strides=4, name='fMaxP2')(convFine)
    convFine = tf.keras.layers.Flatten(name='fFlat1')(convFine)
    
    # network to learn coarse features
    convCoarse = tf.keras.layers.Conv1D(filters=32, kernel_size=Fs*4, strides=int(Fs/2), padding='same', activation='relu', name='cConv1')(inputLayer)
    convCoarse = tf.keras.layers.MaxPool1D(pool_size=4, strides=4, name='cMaxP1')(convCoarse)
    convCoarse = tf.keras.layers.Dropout(rate=0.5, name='cDrop1')(convCoarse)
    convCoarse = tf.keras.layers.Conv1D(filters=128, kernel_size=6, padding='same', activation='relu', name='cConv2')(convCoarse)
    convCoarse = tf.keras.layers.Conv1D(filters=128, kernel_size=6, padding='same', activation='relu', name='cConv3')(convCoarse)
    convCoarse = tf.keras.layers.Conv1D(filters=128, kernel_size=6, padding='same', activation='relu', name='cConv4')(convCoarse)
    convCoarse = tf.keras.layers.MaxPool1D(pool_size=2, strides=2, name='cMaxP2')(convCoarse)
    convCoarse = tf.keras.layers.Flatten(name='cFlat1')(convCoarse)
    
    # concatenate coarse and fine cnns
    mergeLayer = tf.keras.layers.concatenate([convFine, convCoarse], name='merge_1')
    outLayer = tf.keras.layers.Dropout(rate=0.5, name='mDrop1')(mergeLayer)
    
    outLayer = tf.keras.layers.Reshape((1, outLayer.get_shape()[1]), name='reshape1')(outLayer)
    outLayer = tf.keras.layers.LSTM(64, return_sequences=True)(outLayer)
    outLayer = tf.keras.layers.LSTM(64, return_sequences=False)(outLayer)

    # Classify
    outLayer = tf.keras.layers.Dense(n_classes, activation='softmax', name='outLayer')(outLayer)
    model = tf.keras.Model(inputLayer, outLayer)
    optimizer = keras.optimizers.Adam(lr=1e-4)
    model.compile(optimizer=optimizer, loss='categorical_crossentropy', metrics=['acc'])
    model.summary()
    return model


