from models import get_model_cnn
import numpy as np
import tensorflow as tf
from utils import gen, chunker, WINDOW_SIZE, rescale_array
from sklearn.metrics import f1_score, accuracy_score, classification_report
from glob import glob
import os
from sklearn.model_selection import train_test_split
from tqdm import tqdm

def get_input_shape(test_dict):
    for record in tqdm(test_dict):
        all_rows = test_dict[record]['x']

def compute_pruned_count(percent_pruning, layer):
    total = layer.output_shape[2]
    nb_to_be_pruned = int(np.floor(percent_pruning * total))
    return nb_to_be_pruned

def get_weights_l1(layer):
    weights = layer.get_weights()[0]
    norms = []
    for i in range(weights.shape[0]):
        l1_norm = np.sum(abs(weights[i]))
        norms.append(l1_norm)
    return np.array(norms)

def set_weights_to_zero(model, layer, layer_id, perc):  
    weights_norm = get_weights_l1(layer)
    weights = layer.get_weights()[0] 
    n_pruned = compute_pruned_count(perc, layer)
    low_norm_weights = weights_norm.argsort()[:n_pruned]

    for i in range(0, low_norm_weights.shape[0]):
        x = low_norm_weights[i]
        weights[x] = 0
    
    model.layers[layer_id].set_weights([weights])
    model.compile(tf.keras.optimizers.Adam(0.001), tf.keras.losses.sparse_categorical_crossentropy, metrics=['acc'])  
    return model

def test_model(test_dict, model):  
    preds = []
    gt = []
    for record in tqdm(test_dict):
        all_rows = test_dict[record]['x']
        for batch_hyp in chunker(range(all_rows.shape[0])):
            X = all_rows[min(batch_hyp):max(batch_hyp)+1, ...]
            Y = test_dict[record]['y'][min(batch_hyp):max(batch_hyp)+1]     
            X = np.expand_dims(X, 0)
            X = rescale_array(X)
            Y_pred = model.predict(X)
            Y_pred = Y_pred.argmax(axis=-1).ravel().tolist()
            gt += Y.ravel().tolist()
            preds += Y_pred
            
    f1 = f1_score(gt, preds, average="macro")
    acc = accuracy_score(gt, preds)
    print(classification_report(gt, preds))


def main(dataset_path, pretrained_model_path):
    files = sorted(glob(os.path.join(dataset_path, "*.npz")))
    ids = sorted(list(set([x.split("/")[-1][:5] for x in files])))

    #split by test subject
    train_ids, test_ids = train_test_split(ids, test_size=0.15, random_state=1338)
    train_val, test = [x for x in files if x.split("/")[-1][:5] in train_ids],\
                    [x for x in files if x.split("/")[-1][:5] in test_ids]

    train, val = train_test_split(train_val, test_size=0.1, random_state=1337)
    train_data_size = len(train)

    train_dict = {k: np.load(k) for k in train}
    test_dict = {k: np.load(k) for k in test}
    val_dict = {k: np.load(k) for k in val}

    model = get_model_cnn(nclass=5)
    model.load_weights(pretrained_model_path)
    model.compile(tf.keras.optimizers.Adam(0.001), tf.keras.losses.sparse_categorical_crossentropy, metrics=['acc'])

    layer_id = 1
    for j in range(0, 10, 3):
        perc = j/10
        model_pruned = set_weights_to_zero(model, model.layers[layer_id], layer_id, perc)
        test_model(test_dict, model_pruned)
