import numpy as np
import tensorflow as tf
import os
from tensorflow.keras.layers import Flatten, Dense, Conv2D, BatchNormalization, MaxPooling2D
from keras.models import Sequential
from SamplingLayer import SamplingLayer as SamplingLayer 
from keras.models import load_model  
from keras.layers import Conv2D, Dense, Flatten, MaxPooling2D, Permute,Activation,  GlobalAveragePooling2D, Dropout,BatchNormalization, Add
from keras.callbacks import ModelCheckpoint, ReduceLROnPlateau, EarlyStopping
from SamplingLayer import SamplingLayer as SamplingLayer 
from LogMelSpectrogram import LogMelSpectrogram as LogMelSpectrogram
from audiomentations import Compose, AddGaussianNoise, TimeStretch
from keras.models import load_model
import pandas as pd

sr = 44100

def read_ESC_data(esc50_csv, base_data_path):
    pd_data = pd.read_csv(esc50_csv)
    filtered_pd = pd_data.loc[pd_data.esc10]

    classes = filtered_pd.iloc[:, 3].unique()
   
    my_classes = classes
    i = 0
    map_class_to_id = {}
    for c in my_classes:
        map_class_to_id[c] = i
        i = i+1
   
    nb_classes = i

    class_id = filtered_pd['category'].apply(lambda name: map_class_to_id[name])
    filtered_pd = filtered_pd.assign(target=class_id)

    full_path = filtered_pd['filename'].apply(lambda row: os.path.join(base_data_path, row))
    filtered_pd = filtered_pd.assign(filename=full_path)
    
    #Load the audio files
    filenames = filtered_pd['filename']
    targets = filtered_pd['target']
    folds = filtered_pd['fold']
    targets = tf.one_hot(targets, nb_classes)

    main_ds = tf.data.Dataset.from_tensor_slices((filenames, targets, folds))
    main_ds.element_spec

    main_ds = main_ds.map(load_wav_for_map)
    main_ds.element_spec

    #split the data
    cached_ds = main_ds.cache()
    train_ds = cached_ds.filter(lambda embedding, label, fold: fold < 4)
    val_ds = cached_ds.filter(lambda embedding, label, fold: fold == 4)
    test_ds = cached_ds.filter(lambda embedding, label, fold: fold == 5)

    # remove the folds column now that it's not needed anymore
    remove_fold_column = lambda embedding, label, fold: (embedding, label)

    train_ds = train_ds.map(remove_fold_column)
    val_ds = val_ds.map(remove_fold_column)
    test_ds = test_ds.map(remove_fold_column)

    augumented_train_ds = augment_audio_dataset(train_ds)
    combined_dataset = train_ds.concatenate(augumented_train_ds)

    train_ds = combined_dataset
    train_ds = train_ds.cache().shuffle(1000).batch(1).prefetch(tf.data.AUTOTUNE)
    val_ds = val_ds.cache().batch(1).prefetch(tf.data.AUTOTUNE)
    test_ds = test_ds.cache().batch(1).prefetch(tf.data.AUTOTUNE)

    num_labels = len(my_classes)

    return train_ds, val_ds, test_ds, num_labels

def get_size_dataset(train_ds):
  train_ds1 = train_ds.as_numpy_iterator()
  train_ds1 = list(train_ds1)
  return len(train_ds1)

@tf.function
def load_wav_mono(filename):
    file_contents = tf.io.read_file(filename)
    wav, sample_rate = tf.audio.decode_wav(
          file_contents,
          desired_channels=1)
    wav = tf.squeeze(wav, axis=-1)
    return wav

def load_wav_for_map(filename, label, fold):
  return load_wav_mono(filename), label, fold

def apply_pipeline(y, sr):
    augmentations_pipeline = Compose(
    [
        AddGaussianNoise(min_amplitude=0.01, max_amplitude=0.15, p=0),
        TimeStretch(min_rate=.75, max_rate=1.25,p=1),
    ]
    )
    shifted = augmentations_pipeline(y, sr)
    return shifted

@tf.function
def tf_apply_pipeline(audio, label):
    augmented_feature = tf.numpy_function(apply_pipeline, inp=[audio, sr], Tout=tf.float32, name="apply_pipeline")
    return augmented_feature, label

def augment_audio_dataset(dataset: tf.data.Dataset):
    dataset = dataset.map(tf_apply_pipeline)
    return dataset

def define_inference_subnet(input_shape=(220500,), num_labels=None):
    model = tf.keras.models.Sequential()
    model.add(LogMelSpectrogram(sample_rate=44100, fft_size=2048, hop_size=1014, n_mels=128, f_min=0.0, f_max=None))
    model.add(tf.keras.layers.Permute((2,1,3)))
    model.add(BatchNormalization(axis=1, input_shape=input_shape))

    model.add(Conv2D(64, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Conv2D(64, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Conv2D(128, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Conv2D(128, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Conv2D(256, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Conv2D(256, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Flatten())
    model.add(Dense(256, activation = 'relu'))
    model.add(Dropout(0.5))
    model.add(Dense(num_labels, activation="softmax"))

    # compile the model
    model.compile(optimizer='adam',
                    loss='categorical_crossentropy',
                    metrics=['accuracy'])
    #model.summary()
    return model

def compute_pruned_count(percent_pruning, layer):
    total = layer.output_shape[1]
    nb_to_be_pruned = int(np.floor(percent_pruning * total))
    return nb_to_be_pruned

def set_weights_to_zero(model, layer, layer_id, perc):  
    weights = layer.get_weights()[0] 
    weights_norm = abs(weights)
    n_pruned = compute_pruned_count(perc, layer)
    low_norm_weights = weights_norm.argsort()[:n_pruned]
    weights[low_norm_weights] = 0  
    model.layers[layer_id].set_weights([weights])
    return model

def define_sampling_subnet(dim = 220500, input_shape=(220500,),):
    model = tf.keras.models.Sequential()
    model.add(tf.keras.layers.Input(shape=input_shape))
    model.add(SamplingLayer(input_dim=dim))
    model.compile(optimizer='adam',
                    loss='categorical_crossentropy',
                    metrics=['accuracy'])
    return model

def create_model(dim = 220500, input_shape=(220500,), num_labels=None):
    model = tf.keras.models.Sequential()

    model.add(tf.keras.layers.Input(shape=input_shape))

    model.add(SamplingLayer(input_dim=dim))
    model.add(LogMelSpectrogram(sample_rate=44100, fft_size=2048, hop_size=1014, n_mels=128, f_min=0.0, f_max=None))
    model.add(tf.keras.layers.Permute((2,1,3)))
    model.add(BatchNormalization(axis=1, input_shape=input_shape))

    model.add(Conv2D(64, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Conv2D(128, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Conv2D(256, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Flatten())
    model.add(Dense(256, activation = 'relu'))
    model.add(Dropout(0.5))
    model.add(Dense(num_labels, activation="softmax"))

    # compile the model
    model.compile(optimizer='adam',
                    loss='categorical_crossentropy',
                    metrics=['accuracy'])

    return model

def assemble_sparse_model(perc, filename, sampling_subnet_path, inference_subnet_path):
    train_ds, val_ds, test_ds, num_labels = read_ESC_data()
    model = create_model(dim = 220500, input_shape=(220500,), num_labels=num_labels)
    
    subsampling_model = define_sampling_subnet()
    saved_model =  tf.keras.models.load_model(sampling_subnet_path) 
    weights = saved_model.layers[0].get_weights()[0]
    l = []
    l.append(weights)
    subsampling_model.layers[0].set_weights(l)    
    subsampling_model = set_weights_to_zero(subsampling_model, subsampling_model.layers[0], 0, perc) 

    sampling_weights = subsampling_model.layers[0].get_weights()[0]  
    l = []
    l.append(sampling_weights)
    model.layers[0].set_weights(l) 

    inference_model = define_inference_subnet(input_shape=(220500,), num_labels=num_labels)
    opt = tf.keras.optimizers.Adam(learning_rate=0.001)
    inference_model.compile(loss='categorical_crossentropy',
                optimizer=opt,
                metrics=['acc'])

    inference_model = load_model(inference_subnet_path)

    i = 2
    for j in range(1,len(inference_model.layers)):
        model.layers[i].set_weights(inference_model.layers[j].get_weights())
        i = i+1

    accuracy = model.evaluate(test_ds)
    path = os.path.abspath(filename)
    model.save_weights(path)
    model.save(path)

def main():
    perc = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    for p in perc:
        filename = 'model_new'+str(p)
        assemble_sparse_model(p, filename)

