import os.path as os
from keras.callbacks import ReduceLROnPlateau
import tensorflow as tf
from tensorflow.keras.layers import Dropout, Flatten, Dense, Conv2D, Reshape, BatchNormalization, MaxPooling2D, Activation
from SamplingLayer import SamplingLayer as SamplingLayer
from tensorflow.keras import optimizers
from tensorflow.keras.datasets import mnist
from keras.utils.np_utils import to_categorical  
from keras.models import Sequential
import os

learningrate = 0.0002

def createModel_MNIST(in_shape=(28,28), learningrate=0.0002):
    model = Sequential()
    model.add(tf.keras.layers.Input(shape=in_shape))
    model.add(SamplingLayer(input_dim=(28,28)))
    model.add(Reshape(target_shape=(28,28,1)))
    model.add(Conv2D(32, (3, 3), input_shape=(28,28,1)))
    model.add(Activation('relu'))
    BatchNormalization(axis=-1)
    model.add(Conv2D(32, (3, 3)))
    model.add(Activation('relu'))
    model.add(MaxPooling2D(pool_size=(2,2)))
    BatchNormalization(axis=-1)
    model.add(Conv2D(64,(3, 3)))
    model.add(Activation('relu'))
    BatchNormalization(axis=-1)
    model.add(Conv2D(64, (3, 3)))
    model.add(Activation('relu'))
    model.add(MaxPooling2D(pool_size=(2,2)))
    model.add(Flatten())
    BatchNormalization()
    model.add(Dense(512))
    model.add(Activation('relu'))
    BatchNormalization()
    model.add(Dropout(0.2))
    model.add(Dense(10))
    model.add(Activation('softmax'))
    opt = optimizers.Adam(learning_rate=learningrate, beta_1=0.5)
    model.compile(loss='categorical_crossentropy', optimizer=opt, metrics=['accuracy'])

    return model

def train_CSDL_model_MNIST(path_pretrained_model):
    (X_train, y_train), (X_test, y_test) = mnist.load_data()

    X_train = X_train.astype('float32')
    X_test = X_test.astype('float32')

    y_train = to_categorical(y_train, num_classes=10)
    y_test = to_categorical(y_test, num_classes=10)

    X_train = X_train.astype('float32')
    X_test = X_test.astype('float32')

    # Normalization of pixel values (to [0-1] range)
    X_train /= 255
    X_test /= 255

    loss = 'categorical_crossentropy' 
    metrics = ['accuracy']

    model = createModel_MNIST(in_shape=(28,28),learningrate=0.0002) 
    optimizer = optimizers.Adam(lr=learningrate, beta_1=0.5)
    model.compile(optimizer=optimizer, loss=loss, metrics=metrics)

    savedir = os.path.join(os.path.dirname(__file__),'v2')
    if savedir:
        if not os.path.isdir(savedir):
            os.makedirs(savedir)
            
    callbacks = [
            ReduceLROnPlateau(monitor='val_loss', factor=0.5, patience=10, min_lr=1e-7, cooldown=20,verbose=1),
            tf.keras.callbacks.ModelCheckpoint(os.path.join(savedir, 'newweights-{epoch:02d}-{val_loss:.2f}.hdf5'), monitor='val_loss', verbose=1, save_weights_only=True, mode='auto', period=1),
            tf.keras.callbacks.EarlyStopping(monitor="val_loss", patience=10, verbose=1)
            ]

    history = model.fit(X_train,y_train, epochs=30, batch_size=32, callbacks=callbacks, verbose=2, validation_split=0.2)
    loss, accuracy = model.evaluate(X_test, y_test, verbose=0)
    model.save_weights(path_pretrained_model)

