import os
import pathlib
import matplotlib.pyplot as plt
import numpy as np
import math
import tensorflow as tf
from speech_functions import load_data_batches, get_size_dataset
import pandas as pd
from tensorflow.keras.callbacks import LearningRateScheduler
from sklearn.model_selection import train_test_split
from models import define_inference_subnet


seed = 42
tf.random.set_seed(seed)
np.random.seed(seed)

def load_inference_trained():
    inference_model = define_inference_subnet()  
    inference_model.compile(
            optimizer=tf.keras.optimizers.Adam(learning_rate=0.001),
            loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=False),
            metrics=['sparse_categorical_accuracy'], 
        )    
    inference_model.load_weights('inference_subnet_chk.h5')
    path = os.path.abspath('inference_subnet')
    inference_model.save_weights(path)
    inference_model.save("inference_subnet")

def main():
    # Load data
    train_ds, val_ds, test_ds, test_ds_noise, input_shape, wave_orig, commands = load_data_batches()
    train_data_size = get_size_dataset(train_ds)
    num_labels = len(commands)

    inference_model = define_inference_subnet()  
    inference_model.compile(
        optimizer=tf.keras.optimizers.Adam(learning_rate=0.001),
        loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=False),
        metrics=['sparse_categorical_accuracy'], 
    )    

    dataframe2 = pd.read_csv('labels.csv')
    dataframe2 = dataframe2.replace(np.nan, '', regex=True)
    dataframe2 = dataframe2[1:]
    y = tf.convert_to_tensor(dataframe2)
    
    dataframe1 = pd.read_csv('samples.csv')
    dataframe1 = dataframe1.replace(np.nan, '', regex=True)
    dataframe1 = dataframe1[1:]
    x = tf.convert_to_tensor(dataframe1, dtype=np.float16)

    trainX, testX, trainY, testY = train_test_split(np.array(x, dtype=np.float32), np.array(y, dtype=np.float32), test_size=0.3, shuffle=True)

    early_stopping_monitor = tf.keras.callbacks.EarlyStopping(
        monitor='val_sparse_categorical_accuracy',
        min_delta=0,
        patience=10,
        verbose=0,
        mode='max',
        baseline=None,
        restore_best_weights=True
    )
    model_checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(
        filepath='inference_subnet_chk.h5',
        save_weights_only=True,
        monitor='val_sparse_categorical_accuracy',
        mode='max',
        save_best_only=True)

    def step_decay(epoch):
        initial_lrate = 0.001
        drop = 0.4
        epochs_drop = 15.0
        lrate = initial_lrate * math.pow(drop, math.floor((1+epoch)/epochs_drop))
        
        if (lrate < 4e-5):
            lrate = 4e-5
        
        print('Changing learning rate to {}'.format(lrate))
        return lrate

    learning_rate = LearningRateScheduler(step_decay)
    history = inference_model.fit(trainX, trainY, batch_size=64, shuffle=True, epochs=40, validation_split=0.3,  callbacks=[model_checkpoint_callback, early_stopping_monitor, learning_rate])
    
    plt.close('all')
    metrics = history.history
    plt.plot(history.epoch, metrics['loss'], metrics['val_loss'])
    plt.legend(['loss', 'val_loss'])
    plt.show()
    plt.savefig("Retraining_loss_attention_sampling")

    inference_model.load_weights('inference_subnet_chk.h5')
    path = os.path.abspath('inference_subnet')
    inference_model.save_weights(path)
    inference_model.save("inference_subnet")

